from collections import defaultdict
from decimal import Decimal
from decimal import InvalidOperation as DecimalInvalidOperation
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from .exceptions import (
    DuplicateElementError,
    MustError,
    MustTypeError,
    NoElementError,
    NotFoundError,
    NotOneElementError,
)
from .op import MAP_ASSIGN_OP, TH_ASSIGN_OP, TH_ROUND

if TYPE_CHECKING:
    from _typeshed import SupportsKeysAndGetItem


T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


class EmptyLambda:
    def __call__(self, x):
        return x

    def __str__(self):
        return "lambda x: x"


lambda_empty = EmptyLambda()
# lambda_empty = lambda x: x
# lambda_empty.__str__ = lambda: "lambda x: x"


def name_as(name):
    def wrapper(func):
        func.__name__ = name
        return func

    return wrapper


marked = []


def mark(func):
    marked.append(func)
    return func


def name_as(name):
    def wrapper(func):
        func.__name__ = name
        return func

    return wrapper


@name_as("list")
def __list(iterable: Iterable[T]) -> List[T]:
    return list(iterable)


@overload
def __dict(mapping: "SupportsKeysAndGetItem[K, V]") -> Dict[K, V]:
    return Dict[K, V](mapping)


@overload
def __dict(iterable: Iterable[Tuple[K, V]]) -> Dict[K, V]:
    return Dict[K, V](iterable)


@name_as("dict")
def __dict(iterable):
    return Dict[K, V](iterable)


###########################################
# generator
###########################################
@mark
@name_as("iter")
def __iter(self, selector=None):
    """イテラブルまたはマッピングからイテレータを取得します。
    マッピングの場合は、キーバリューのタプルを返すイテレータを取得します。

    Args:

    * self: イテレータを取得するイテラブルまたはマッピング
    * selector: 取得したイテレータから任意の要素を抽出する関数

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.iter([1, 2]).to(list)
    [1, 2]
    >>> pnq.iter([{"id": 1, "name": "bob"}]).to(list)
    [("id", 1), ("name", "bob")]
    >>> pnq.iter([{"id": 1, "name": "bob"}], lambda x: x[1]).to(list)
    [1, "bob"]
    ```
    """

    it = getattr(self, "__piter__", None)
    if it:
        it = it()
    else:
        if isinstance(self, Mapping):
            it = self.items()
        else:
            it = self

    return __map_nullable(it, selector)


def __map_nullable(self, selector):
    if selector is None:
        return self
    else:
        return map(selector, self)


@mark
def __next(self):
    raise NotImplementedError()


@mark
def value(*args, **kwargs):
    """１つの要素を返すイテレータを生成します。
    位置引数を与えた場合は与えた値かタプルになり、キーワード引数を与えた場合は辞書になります。

    Usage:
    ```
    >>> pnq.value(1).to(list)
    [1]
    >>> pnq.value("a", "b").to(list)
    [("a", "b")]
    >>> pnq.value(name="test").to(list)
    [{"naem": "test"}]
    ```
    """
    if args and kwargs:
        raise ValueError("value() can't accept both positional and keyword arguments")

    if kwargs:
        val = kwargs

    elif len(args) > 1:
        val = args

    elif len(args) == 1:
        val = args[0]

    else:
        raise NotImplementedError()

    yield val


@mark
def infinite(func, *args, **kwargs):
    """渡した関数を無限に実行するイテレータを生成します。
    無限に繰り返されるため、`take`等で終了条件を設定するように注意してください。

    Args:

    * func(args, kwargs): 無限に実行する関数
    * args: 関数に渡す位置引数
    * kwargs: 関数に渡すキーワード引数

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.infinite(datetime.now).take(1).to(list)
    [datetime.datetime(2021, 9, 10, 3, 57, 54, 402467)]
    >>> pnq.infinite(datetime, 2020, 1, day=2).take(1).to(list)
    [datetime.datetime(2010, 1, 2, 0, 0)]
    ```
    """
    while True:
        yield func(*args, **kwargs)


@mark
def repeat(value):
    """同じ値を繰り返す無限イテレータを生成します。

    Args:

    * value: 繰り返す値

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.repeat(5).take(3).to(list)
    [5, 5, 5]
    ```
    """
    from itertools import repeat

    yield from repeat(value)


@mark
def count(start=0, step=1):
    """連続した値を無限に返すイテレータを生成します。
    無限に繰り返されるため、`take`等で終了条件を設定するように注意してください。

    Args:

    * start: 開始値
    * step: 増分

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.count().take(3).to(list)
    [0, 1, 2]
    >>> pnq.count(1, 2).take(3).to(list)
    [1, 3, 5]
    ```
    """
    from itertools import count

    yield from count(start, step)


@mark
def cycle(iterable, repeat=None):
    """イテラブルが返す値を無限に繰り返すイテレータを生成します。

    Args:

    * iterable: 繰り返すイテラブル
    * repeat: 繰り返す回数。Noneの場合は無限に繰り返します。

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.cycle([1,2,3]).take(4).to(list)
    [1, 2, 3, 1]
    >>> pnq.cycle([1,2,3], repeat=2).to(list)
    [1, 2, 3, 1, 2, 3]
    ```
    """
    from itertools import cycle

    yield from cycle(iterable, repeat)


@mark
@name_as("range")
def __range(*args, **kwargs):
    """指定した開始数と終了数までの連続した値を返すイテレータを生成します。

    Args:

    * stop: 終了数

    Args:

    * start: 開始数
    * stop: 終了数
    * step: 増分

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.range(5).to(list)
    [0, 1, 2, 3, 4]
    >>> pnq.range(0, 3, 2).to(list)
    [0, 2, 4]
    ```
    """
    yield from range(*args, **kwargs)


###########################################
# mapping
###########################################


@mark
@name_as("map")
def __map(self, selector):
    """シーケンスの各要素を新しいフォームに射影します。
    str関数を渡した場合、利便性のため`None`は`""`を返します（Pythonの標準動作は`"None"`を返します）。

    Args:

    * self: 変換対象のシーケンス
    * selector(x): 各要素に対する変換関数

    Returns: 変換関数で得られた要素を含むクエリ

    Usage:
    ```
    >>> pnq.query([1]).map(lambda x: x * 2).to(list)
    [2]
    >>> pnq.query([None]).map(str).to(list)
    [""]
    ```
    """
    if selector is str:
        selector = lambda x: "" if x is None else str(x)
    return map(selector, self)


def __map_nullable(self, selector):
    if selector is None:
        return self
    else:
        return map(selector, self)


def itertools__():
    from itertools import (
        chain,
        combinations,
        combinations_with_replacement,
        count,
        cycle,
        dropwhile,
        groupby,
        islice,
        permutations,
        product,
        repeat,
        starmap,
        takewhile,
        tee,
        zip_fillvalue,
        zip_longest,
    )


@mark
def select_single_node(
    self,
    node_selector=lambda x: x.node,
    return_selector=lambda nest, parent, child: (nest, parent, child),
):
    pass


@mark
def unpack_pos(self, selector):
    """シーケンスの各要素をアンパックし、新しいフォームに射影します。

    Args:

    * self: 変換対象のシーケンス
    * selector: 各要素に対する変換関数

    Returns: 変換関数で得られた要素を返すクエリ

    Usage:
    ```
    >>> pnq.query([(1, 2)]).unpack_pos(lambda arg1, arg2: arg1)).to(list)
    [1]
    >>> pnq.query([(1, 2, 3, 4, 5)]).unpack_pos(lambda arg1, arg2, *args: args).to(list)
    [(3, 4, 5)]
    ```
    """
    for elm in self:
        yield selector(*elm)


starmap = unpack_pos


@mark
def unpack_kw(self, selector):
    """シーケンスの各要素をキーワードアンパックし、新しいフォームに射影します。

    Args:

    * self: 変換対象のシーケンス
    * selector(kwargs): 各要素に対する変換関数

    Returns: 変換関数で得られた要素を返すクエリ

    Usage:
    ```
    >>> pnq.query([{"id": 1, "name": "bob"}]).unpack_kw(lambda id, name: name)).to(list)
    ["bob"]
    >>> pnq.query([{"id": 1, "name": "bob", "age": 20}]).unpack_kw(lambda id, name, **kwargs: kwargs)).to(list)
    [{"age": 20}]
    ```
    """
    for elm in self:
        yield selector(**elm)  # type: ignore


@mark
def select(self, field, *fields, attr: bool = False):
    """シーケンスの各要素からアイテムを選択し新しいフォームに射影します。
    複数のアイテムを選択した場合は、タプルとして射影します。

    Args:

    * self: 変換対象のシーケンス
    * field: 各要素から選択するアイテム
    * fields: 各要素から選択する追加のアイテム
    * attr: 要素の属性から取得する場合はTrue

    Returns: 選択したアイテムまたは複数のアイテム（タプル）を返すクエリ

    Usage:
    ```
    >>> pnq.query([(1, 2)]).select(0).to(list)
    [1]
    >>> pnq.query([{"id": 1, "name": "a"}]).select("id", "name").to(list)
    [(1, "a")]
    >>> pnq.query([user]).select("id", "name", attr=True).to(list)
    [(1, "a")]
    ```
    """
    pass


@mark
def select_as_tuple(self, *fields, attr: bool = False):
    """シーケンスの各要素からアイテムまたは属性を選択し辞書として新しいフォームに射影します。
    selectと似ていますが、選択した値が１つでも必ずタプルを返します。

    Args:

    * self: 変換対象のシーケンス
    * fields: 選択するアイテムまたは属性
    * attr: 属性から取得する場合はTrueとする

    Returns: 選択したアイテムを含む辞書を返すクエリ

    Usage:
    ```
    >>> pnq.query([(1, 2)]).select_as_tuple(0).to(list)
    [(1,)]
    >>> pnq.query([user]).select_as_tuple("id", "name", attr=True).to(list)
    [("1", "a")]
    ```
    """
    pass


@mark
def select_as_dict(self, *fields, attr: bool = False):
    """シーケンスの各要素からアイテムまたは属性を選択し辞書として新しいフォームに射影します。

    Args:

    * self: 変換対象のシーケンス
    * fields: 選択するアイテムまたは属性
    * attr: 属性から取得する場合はTrueとする

    Returns: 選択したアイテムを含む辞書を返すクエリ

    Usage:
    ```
    >>> pnq.query([(1, 2)]).select_as_dict(0).to(list)
    [{0: 1}]
    >>> pnq.query([user]).select_as_dict("id", "name", attr=True).to(list)
    [{"id": 1, "name": "b"}]
    ```
    """
    pass


@mark
def reflect(self, mapping, *, default=NoReturn, attr: bool = False):
    """シーケンスの各要素のフィールドを与えたマッピングに基づき辞書として新しいフォームに射影します。

    Args:

    * self: 変換対象のシーケンス
    * mapping: 元の要素のフィールドと射影先のフィールドの対応表
    * default: フィールドを取得できない場合のデフォルト値
    * attr: 属性から取得する場合はTrueとする

    Returns: 参照したシーケンスをそのまま返す

    Usage:
    ```
    >>> person = {"id":1, "name": "山田", "kana": "やまだ", "note": "hoge"}
    >>> pnq.query([person]).reflect({
    >>>   "id": "id",
    >>>   "name": {"name", "searchable"},
    >>>   "kana": {"kana", "searchable"},
    >>>   "note": "searchable"
    >>> }).to(list)
    >>> [{"id": 1, "name": "山田", "kana": "やまだ", "searchable": ["山田", "やまだ", "hoge"]}]
    ```
    """


@mark
def flat(self, selector=None):
    """シーケンスの各要素をイテラブルに射影し、その結果を１つのシーケンスに平坦化します。

    Args:

    * self: 変換対象のシーケンス
    * selector: 各要素から平坦化する要素を選択する関数

    Returns: 参照したシーケンスをそのまま返す

    Usage:
    ```
    >>> pnq.query(["abc", "def"]).flat().to(list)
    >>> ["a", "b", "c", "d", "e", "f"]
    >>> countries = [{"country": "japan", "state": ["tokyo", "osaka"]}, {"country": "america", "state": ["new york", "florida"]}]
    >>> pnq.query(countries).flat(lambda x: x["state"]).to(list)
    >>> ["tokyo", "osaka", "new york", "florida"]
    ```
    """


select_many = flat
flat_map = flat


@mark
def flat_recursive(self, selector):
    """シーケンスの各要素から再帰的に複数ノードを選択し、選択されたノードを１つのシーケンスに平坦化します。
    各ルート要素から浅い順に列挙されます。

    Args:

    * self: 変換対象のシーケンス
    * selector: 各要素から平坦化する要素を再帰的に選択する関数（戻り値はリスト等に含めて返す必要があります）

    Returns: 参照したシーケンスをそのまま返す

    Usage:
    ```
    >>> pnq.query(
    >>>     {"name": "a", "nodes": [{"name": "b", nodes: [{"name": c, "nodes": []}, {"name": "d", "nodes": []}}}]}]}
    >>> ).flat_recursive(lambdax x: x["nodes"]).select("name").to(list)
    >>> ["a", "b", "c", "d"]
    ```
    """


def pivot_unstack(self, default=None):
    """行方向に並んでいるデータを列方向に入れ替える

    Args:

    * self: 変換対象のシーケンス
    * default: フィールドが存在しない場合のデフォルト値

    Returns: 参照したシーケンスをそのまま返す

    Usage:
    ```
    data = [
        {"name": "test1", "age": 20},
        {"name": "test2", "age": 25},
        {"name": "test3", "age": 30, "sex": "male"},
    ]
    {'name': ['test1', 'test2', 'test3'], 'age': [20, 25, 30], 'sex': [None, None, 'male']}
    ```
    """
    from collections import defaultdict

    dataframe = {}
    data = []

    # 全てのカラムを取得
    for i, dic in enumerate(self):
        data.append(dic)
        for k in dic.keys():
            dataframe[k] = None

    # カラム分の領域を初期化
    for k in dataframe:
        dataframe[k] = []

    # データをデータフレームに収める
    for dic in data:
        for k in dataframe.keys():
            v = dic.get(k, default)
            dataframe[k].append(v)

    yield from dataframe.items()


def pivot_stack(self):
    """列方向に並んでいるデータを行方向に入れ替える

    Args:

    * self: 変換対象のシーケンス

    Returns: 参照したシーケンスをそのまま返す

    Usage:
    ```
    {'name': ['test1', 'test2', 'test3'], 'age': [20, 25, 30], 'sex': [None, None, 'male']}
    data = [
        {"name": "test1", "age": 20, "sex": None},
        {"name": "test2", "age": 25, "sex": None},
        {"name": "test3", "age": 30, "sex": "male"},
    ]
    ```
    """
    data = dict(self)
    columns = list(data.keys())

    for i in range(len(columns)):
        row = {}
        for c in columns:
            row[c] = data[c][i]

        yield row


@mark
def cast(self, type):
    """シーケンスの型注釈を変更します。この関数はエディタの型解釈を助けるためだけに存在し、何も処理を行いません。

    実際に型を変更する場合は、`map`を使用してください。

    Args:

    * self: 変換対象のシーケンス
    * type: 新しい型注釈

    Returns: 参照したシーケンスをそのまま返す

    Usage:
    ```
    >>> pnq.query([1]).cast(float)
    ```
    """
    pass


@mark
@name_as("enumerate")
def __enumerate(self, start: int = 0, step: int = 1):
    """シーケンスの各要素とインデックスを新しいフォームに射影します。

    Args:

    * self: 変換対象のシーケンス
    * start: 開始インデックス
    * step: 増分

    Returns: インデックスと要素（タプル）を返すクエリ

    Usage:
    ```
    >>> pnq.query([1, 2]).enumerate().to(list)
    [(0, 1), (1, 2)]
    >>> pnq.query([1, 2]).enumerate(5).to(list)
    [(5, 1), (6, 2)]
    >>> pnq.query([1, 2]).enumerate(0, 10)).to(list)
    [(0, 1), (10, 2)]
    ```
    """
    pass


@mark
def group_by(self, selector=lambda x: x):
    """シーケンスの各要素からセレクタ関数でキーとバリューを取得し、キーでグループ化されたシーケンスを生成します。
    セレクタ関数を指定しない場合、各要素がすでにキーバリューのタプルであることを期待し、キーでグループ化します。

    Args:

    * self: 変換対象のシーケンス
    * selector: キーとバリューを選択する関数

    Returns: キーと要素（タプル）を返すクエリ

    Usage:
    ```
    >>> data = [
    >>>   {"name": "banana", "color": "yellow", "count": 3},
    >>>   {"name": "apple", "color": "red", "count": 2},
    >>>   {"name": "strawberry", "color": "red", "count": 5},
    >>> ]
    >>> pnq.query(data).group_by(lambda x: x["color"], x["name"]).to(list)
    [("yellow", ["banana"]), ("red", ["apple", "strawberry"])]
    >>> pnq.query(data).select("color", "count").group_by().to(dict)
    {"yellow": [3], "red": [2, 5]}
    ```
    """
    results = defaultdict(list)
    for elm in self:
        k, v = selector(elm)
        results[k].append(v)

    for k, v in results.items():
        yield k, v


@mark
def chunked(self, size: int):
    """
    シーケンスを指定したサイズ毎に分割する。
    """


@mark
def tee(self, size: int):
    """
    ひとつのイテレータから独立したN個のイテレータを返す
    """


@mark
def join(self, right, on, select):
    pass


@mark
def group_join(self, right, on, select):
    pass


@mark
def request(self, func, retry: int = None):
    """シーケンスから流れてくる値を同期関数に送出するように要求します。
    例外はキャッチされ、実行結果を返すイテレータを生成します。
    関数呼び出し時にキーワードアンパックするため、要素は辞書である必要があります。

    Args:

    * self: 辞書を要素とするシーケンス
    * func: 値の送出先の関数

    Returns: 実行結果

    Usage:
    ```
    >>> def do_something(id, val):
    >>>   if val:
    >>>     return 1
    >>>   else:
    >>>     raise ValueError(val)
    >>>
    >>> for res in pnq.query([{"id": 1, "val": True}, {"id": 2, "val": False}]).request(do_something):
    >>>   if res.err:
    >>>     print(f"ERROR: {res.to(dict)}")
    >>>   else:
    >>>     print(f"SUCCESS: {res.to(dict)}")
    ```
    """
    from .requests import Response, StopWatch

    if retry:
        raise NotImplementedError("retry not implemented")

    for v in self:

        with StopWatch() as sw:
            err = None
            result = None
            try:
                result = func(**v)
            except Exception as e:
                err = e

        res = Response(
            func, kwargs=v, err=err, result=result, start=sw.start, end=sw.end
        )

        yield res


async def request_async(self, func, timeout: float = None, retry: int = None):
    """シーケンスから流れてくる値を非同期関数に送出するように要求します。
    例外はキャッチされ、実行結果を返すイテレータを生成します。
    関数呼び出し時にキーワードアンパックするため、要素は辞書である必要があります。

    Args:

    * self: 辞書を要素とするシーケンス
    * func: 値の送出先の関数

    Returns: 実行結果

    Usage:
    ```
    >>> async def do_something(id, val):
    >>>   if val:
    >>>     return 1
    >>>   else:
    >>>     raise ValueError(val)
    >>>
    >>> params = pnq.query([{"id": 1, "val": True}, {"id": 2, "val": False}])
    >>> async for res in params.request_async(do_something):
    >>>   if res.err:
    >>>     print(f"ERROR: {res.to(dict)}")
    >>>   else:
    >>>     print(f"SUCCESS: {res.to(dict)}")
    ```
    """
    from .requests import Response, StopWatch

    if retry:
        raise NotImplementedError("retry not implemented")

    if timeout:
        raise NotImplementedError("timeout not implemented")

    for v in self:

        with StopWatch() as sw:
            err = None
            result = None
            try:
                result = await func(**v)
            except Exception as e:
                err = e

        res = Response(
            func, kwargs=v, err=err, result=result, start=sw.start, end=sw.end
        )

        yield res


def debug(self, breakpoint=lambda x: x, printer=print):
    for v in self:
        printer(v)
        breakpoint(v)
        yield v


###########################################
# Expander set operation
###########################################

# https://docs.aws.amazon.com/ja_jp/redshift/latest/dg/r_UNION.html


@mark
def union_all(self):
    """全ての行を集合に含む"""
    pass


extend = union_all


@mark
def union(self, *iterables):
    """全ての行を集合に含み重複は許可しない"""
    pass


@mark
def union_intersect(self):
    """共通部分のみ抽出"""
    pass


@mark
def union_minus(self):
    """1つ目の問い合わせには存在するが、2つ目の問い合わせには存在しないデータを抽出
    exceptと同じ意味
    """
    pass


# difference
# symmetric_difference

# < <= > >= inclusion


@mark
@name_as("zip")
def __zip(self, *iterables):
    raise NotImplementedError()


def compress(self, *iterables):
    """未実装"""
    import itertools

    return itertools.compress(self, *iterables)


@mark
def cartesian(self, *iterables):
    """複数のシーケンスのデカルト積を返します。

    Args:

    * self, iterables: デカルト積を求める複数のシーケンス
    """


"""
# product（デカルト積）
from itertools import product
mark=['♠', '♥', '♦', '♣']
suu = ['A', '2', '3', '4', '5', '6', '7', '8', '9', '10', 'J', 'Q', 'K']
tramp = [m for m in product(mark,suu)]
print(tramp)
# [('♠', 'A'), ('♠', '2'), ...]
"""


###########################################
# filtering
###########################################
@mark
@name_as("filter")
def __filter(self, predicate):
    """述語に基づいてシーケンスの要素をフィルタ処理します。

    Args:

    * self: フィルタ対象のシーケンス
    * predicate: 条件を満たすか検証する関数

    Returns: 条件を満たす要素を返すクエリ

    Usage:
    ```
    >>> pnq.query([1, 2]).filter(lambda x: x == 1).to(list)
    [1]
    >>> pnq.query({1: True, 2: False, 3: True}).filter(lambda x: x[1] == True).to(list)
    [(1, True), (3, True)]
    ```
    """
    return filter(predicate, self)


@mark
def must(self, predicate, msg=""):
    """述語に基づいてシーケンスの要素を検証します。
    検証に失敗した場合、即時に例外が発生します。

    Args:

    * self: フィルタ対象のシーケンス
    * predicate: 条件を満たすか検証する関数

    Returns: 全要素を返すクエリ（例外が発生しない限り）

    Usage:
    ```
    >>> pnq.query([1, 2]).must(lambda x: x == 1).to(list)
    raise ValueError("2")
    >>> pnq.query({1: True, 2: False, 3: True}).must(lambda x: x[1] == True).to(list)
    raise ValueError("(2, False)")
    ```
    """
    for elm in self:
        if not predicate(elm):
            raise MustError(f"{msg} {elm}")
        yield elm


@mark
def filter_type(self, *types):
    """指定した型に一致するシーケンスの要素をフィルタ処理します。
    型は複数指定することができ、`isinstance`の挙動に準じます。

    Args:

    * self: フィルタ対象のシーケンス
    * types: フィルタする型

    Returns: 指定した型の要素を返すクエリ

    Usage:
    ```
    >>> pnq.query([1, False, "a"]).filter_type(int).to(list)
    [1, False]
    >>> pnq.query([1, False, "a"]).filter_type(str, bool).to(list)
    [False, "a"]
    ```
    """
    for elm in self:
        if isinstance(elm, *types):
            yield elm


@mark
def must_type(self, types):
    """シーケンスの要素が指定した型のいずれかであるか検証します。
    検証に失敗した場合、即時に例外が発生します。
    型は複数指定することができ、`isinstance`の挙動に準じます。

    Args:

    * self: フィルタ対象のシーケンス
    * types: フィルタする型

    Returns: 全要素を返すクエリ（例外が発生しない限り）

    Usage:
    ```
    >>> pnq.query([1, 2]).must_type(str, int).to(list)
    raise ValueError("1 is not str")
    ```
    """
    for elm in self:
        if not isinstance(elm, types):
            raise MustTypeError(f"{elm} is not {tuple(x.__name__ for x in types)}")
        yield elm


@mark
def filter_keys(self, *keys):
    """シーケンスの要素から指定したキーの要素のみフィルタ処理します。
    このメソッドは、`list` `dict` `set`などをクエリ化した直後のみ利用できます。

    * `list`、`tuple`の場合インデックスでフィルタされ、値を返します。
    * `dict`の場合キーでフィルタされ、キーと要素を返します。
    * `set`の場合は、キーでフィルタされ、キーを返します。

    Args:

    * self: フィルタ対象のシーケンス
    * keys: フィルタするキー

    Returns: 指定したキーの要素のみ返すクエリ

    Usage:
    ```
    >>> pnq.query([1, 2]).filter_keys(1).to(list)
    [2]
    >>> pnq.query({"a": 1, "b": 2}).filter_keys("b").to(list)
    [("b", 2)]
    >>> pnq.query({"a", "b"}).filter_keys("b").to(list)
    ["b"]
    ```
    """


@mark
def must_keys(self, *keys):
    """`filter_keys`を実行し、全てのキーを取得できなかった場合例外を発生させます。
    検証が完了するまで、ストリームは保留されます。
    """

    not_exists = set()
    key_values = []
    raise_if_not_unique_keys(keys)
    undefined = object()

    for k in keys:
        val = get(self, k, undefined)
        if val is undefined:
            not_exists.add(k)
        else:
            key_values.append((k, val))

    if not_exists:
        raise NotFoundError(str(not_exists))

    if typ == "map":
        for k, v in key_values:
            yield k, v
    elif typ == "seq":
        for k, v in key_values:
            yield v
    elif typ == "set":
        for k, v in key_values:
            yield k
    else:
        raise TypeError(f"unknown type: {typ}")


@mark
def filter_unique(self, selector=None):
    """シーケンスの要素から重複する要素を除去する。
    セレクタによって選択された値に対して重複が検証され、その値を返す。

    Args:

    * self: フィルタ対象のシーケンス
    * selector: 重複を検証する値（複数の値を検証する場合はタプル）

    Returns: 重複を含まない要素を返すクエリ

    Usage:
    ```
    >>> pnq.query([1, 2, 1]).filter_unique().to(list)
    [1, 2]
    >>> pnq.query([(0 , 0 , 0), (0 , 1 , 1), (0 , 0 , 2)]).unique(lambda x: (x[0], x[1])).to(list)
    [(0, 0), (0, 1)]
    ```
    """
    duplidate = set()
    for value in __map_nullable(self, selector):
        if value in duplidate:
            pass
        else:
            duplidate.add(value)
            yield value


distinct = filter_unique


@mark
def must_unique(self, selector=None):
    """シーケンスの要素から値を選択し、選択した値が重複していないか検証します。

    Args:

    * self: フィルタ対象のシーケンス
    * selector: 検証する値を選択する関数
    * immediate: 即時に例外を発生させる

    Returns: 全要素を返すクエリ（例外が発生しない限り）

    Usage:
    ```
    >>> pnq.query([1, 2, 1]).must_unique().to(list)
    raise DuplicateError("1")
    ```

    """
    duplidate = set()
    for value in __map_nullable(self, selector):
        if value in duplidate:
            raise DuplicateElementError(value)
        else:
            duplidate.add(value)
            yield value


def raise_if_not_unique_keys(keys):
    if isinstance(keys, set):
        unique = keys
    else:
        unique = set(keys)
    if len(unique) != len(keys):
        raise TypeError(f"can't accept duplicate keys: {keys}")


###########################################
# partitioning
###########################################


@mark
def take(self, count_or_range: int):
    """シーケンスから指定した範囲の要素を返します。

    Args:

    * self: 取得対象のシーケンス
    * count_or_range: シーケンスの先頭から取得する要素数または取得する範囲

    Returns: 取得された要素を返すクエリ

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).take(2).to(list)
    [1, 2]
    >>> pnq.query([1, 2, 3]).take(range(1, 2)).to(list)
    [2]
    ```
    """
    pass


@mark
def take_while(self, predicate):
    """シーケンスの先頭から、条件の検証に失敗するまでの要素を返します。
    **検証に失敗した要素は破棄されるため、値を消費するイテレータがソースの場合は注意が必要です。**

    Args:

    * self: バイパス対象のシーケンス
    * predicate: 条件を検証する関数

    Returns: 取得された要素を返すクエリ

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).enumerate().take_while(lambda v: v[0] < 2).select(1).to(list)
    [1, 2]
    ```
    """
    pass


@mark
def skip(self, count_or_range: int):
    """シーケンスから指定した範囲の要素をバイパスします。

    Args:

    * self: バイパス対象のシーケンス
    * count_or_range: シーケンスの先頭からバイパスする要素数またはバイパスする範囲

    Returns: 取得された要素を返すクエリ

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).skip(1).to(list)
    [2, 3]
    >>> pnq.query([1, 2, 3]).skip(range(1, 2)).to(list)
    [1, 3]
    ```
    """
    pass


@mark
def skip_while(self, predicate):
    """シーケンスの先頭から、条件の検証に失敗するまでの要素をバイパスし、残りの要素を返します。

    Args:

    * self: バイパス対象のシーケンス
    * predicate: 条件を検証する関数

    Returns: 取得された要素を返すクエリ

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).enumerate().skip_while(lambda v: v[0] < 1).select(1).to(list)
    [2, 3]
    ```
    """
    pass


@mark
def take_page(self, page: int, size: int):
    """シーケンスから指定した範囲の要素を返します。
    範囲はページサイズと取得対象のページから求められます。

    Args:

    * self: バイパス対象のシーケンス
    * page: 取得対象のページ（1始まり）
    * size: １ページあたりの要素数

    Returns: 取得された要素を返すクエリ

    Usage:
    ```
    >>> pnq.query([0, 1, 2, 3, 4, 5]).take_page(page=1, size=2).to(list)
    [0, 1]
    >>> pnq.query([0, 1, 2, 3, 4, 5]).take_page(page=2, size=3).to(list)
    [3, 4, 5]
    ```
    """


def take_page_calc(page: int, size: int):
    if size < 0:
        raise ValueError("size must be >= 0")
    start = (page - 1) * size
    stop = start + size
    return start, stop


###########################################
# order
###########################################


@mark
def order_by(self, selector=None, desc: bool = False):
    """シーケンスの要素を昇順でソートします。

    Args:

    * self: ソート対象のシーケンス
    * selector: 要素からキーを抽出する関数。複数のキーを評価する場合は、タプルを返してください。
    * desc: 降順で並べる場合はTrue

    Usage:
    ```
    >>> pnq.query([3, 2, 1]]).order_by().to(list)
    [1, 2, 3]
    >>> pnq.query([1, 2, 3]).order_by(lambda x: -x).to(list)
    [3, 2, 1]
    >>> pnq.query([1, 2, 3]).order_by(desc=True).to(list)
    [3, 2, 1]
    >>> pnq.query([(1, 2)), (2, 2), (2, 1)]).order_by(lambda x: (x[0], x[1])).to(list)
    [(1, 2)), (2, 1), (2, 2)]
    ```
    """
    yield from sorted(self, key=selector, reverse=desc)


@mark
def order_by_fields(self, *fields, desc: bool = False, attr: bool = False):
    raise NotImplementedError()


@mark
def order_by_reverse(self):
    """シーケンスの要素を逆順にします。

    Args:

    * self: ソート対象のシーケンス

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).order_by_reverse().to(list)
    [3, 2, 1]
    ```
    """
    if hasattr(self, "__reversed__"):
        if isinstance(self, Mapping):
            return ((k, self[k]) for k in reversed(self))  # type: ignore
        else:
            return reversed(self)
    else:
        return reversed(list(self))


@mark
def order_by_shuffle(self):
    """シーケンスの要素をランダムにソートします。

    Args:

    * self: ソート対象のシーケンス

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).order_by_shuffle().to(list)
    [1, 3, 2]
    >>> pnq.query([1, 2, 3]).order_by_shuffle().to(list)
    [3, 1, 2]
    ```
    """
    import random

    yield from sorted(self, key=lambda k: random.random())


###########################################
# statistics
###########################################
@mark
@name_as("len")
def __len(self):
    """シーケンスの要素数を返します。

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).len()
    3
    ```
    """
    if hasattr(self, "__len__"):
        return len(self)

    i = 0
    for i, v in enumerate(self, 1):
        ...

    return i


@mark
def exists(self):
    """シーケンス内の要素の有無を確認します。

    Usage:
    ```
    >>> pnq.query([]).exists()
    False
    >>> pnq.query([1]).exists()
    True
    ```
    """
    for elm in self:
        return True

    return False


@mark
@name_as("all")
def __all(self, selector=lambda x: x):
    """シーケンスの全ての要素がTrueと判定できるか評価します。要素がない場合はTrueを返します。

    * selector: 要素から検証する値を抽出する関数

    Usage:
    ```
    >>> pnq.query([]).all()
    True
    >>> pnq.query([0]).all()
    False
    >>> pnq.query([1]).all()
    True
    >>> pnq.query([1, 0]).all()
    False
    >>> pnq.query([1, 2]).all()
    True
    ```
    """
    return all(map(selector, self))


@mark
@name_as("any")
def __any(self, selector=lambda x: x):
    """シーケンスのいずれかの要素がTrueと判定できるか評価します。要素がない場合はFalseを返します。

    * selector: 要素から検証する値を抽出する関数

    Usage:
    ```
    >>> pnq.query([]).any()
    False
    >>> pnq.query([0]).any()
    False
    >>> pnq.query([1]).any()
    True
    >>> pnq.query([1, 0]).any()
    True
    ```
    """
    return any(map(selector, self))


@mark
def contains(self, value, selector=lambda x: x) -> bool:
    """既定の等値比較子を使用して、指定した要素がシーケンスに含まれているか評価します。
    辞書をソースとした場合は、キーバリューのタプルを比較します。


    * value: 検索対象の値
    * selector: 要素から検索する値を抽出する関数

    Usage:
    ```
    >>> fruits = pnq.query(["apple", "orange"])
    >>> fruits.contains("banana")
    False
    >>> fruits.contains("apple")
    True
    >>> fruits.contains("orange")
    True
    >>> pnq.query({"a": 1, "b": 2}).contains("a")
    False
    >>> pnq.query({"a": 1, "b": 2}).contains(("a", 1))
    True
    ```
    """
    for val in map(selector, self):
        if val == value:
            return True

    return False


@mark
@name_as("min")
def __min(self, selector=lambda x: x, default=NoReturn):
    """シーケンスの要素から最小の値を取得します。

    Args:

    * selector: 要素から計算する値を抽出する関数
    * default: 要素が存在しない場合に返す値

    Usage:
    ```
    >>> pnq.query([1, 2]).min()
    1
    >>> pnq.query([]).min()
    ValueError: min() arg is an empty sequence
    >>> pnq.query([]).min(default=0)
    0
    ```
    """
    if default is NoReturn:
        return min(map(selector, self))
    else:
        return min(map(selector, self), default=default)


@mark
@name_as("max")
def __max(self, selector=lambda x: x, default=NoReturn):
    """シーケンスの要素から最大の値を取得します。

    Args:

    * selector: 要素から計算する値を抽出する関数
    * default: 要素が存在しない場合に返す値

    Usage:
    ```
    >>> pnq.query([1, 2]).max()
    2
    >>> pnq.query([]).max()
    ValueError: max() arg is an empty sequence
    >>> pnq.query([]).max(default=0)
    0
    ```
    """
    if default is NoReturn:
        return max(map(selector, self))
    else:
        return max(map(selector, self), default=default)


@mark
@name_as("sum")
def __sum(self, selector=lambda x: x):
    """シーケンスの要素を合計します。

    * selector: 要素から計算する値を抽出する関数

    Usage:
    ```
    >>> pnq.query([]).sum()
    0
    >>> pnq.query([1, 2]).sum()
    3
    ```
    """
    return sum(map(selector, self))


@mark
def average(
    self, selector=lambda x: x, exp: float = 0.00001, round: TH_ROUND = "ROUND_HALF_UP"
) -> Union[float, Decimal]:
    """シーケンスの要素の平均を求めます。

    Args:

    * selector: 要素から計算する値を抽出する関数
    * exp: 丸める小数点以下の桁数
    * round: 丸め方式

    Usage:
    ```
    >>> pnq.query([]).average()
    0
    >>> pnq.query([1, 2]).average()
    1.5
    ```
    """
    # import statistics
    # return statistics.mean(pmap(self, selector))  # type: ignore

    seed = Decimal("0")
    i = 0
    val = 0

    for i, val in enumerate(map(selector, self), 1):
        # val = selector(val)
        try:
            val = Decimal(str(val))  # type: ignore
        except DecimalInvalidOperation:
            raise TypeError(f"{val!r} is not a number")
        seed += val

    if i:
        result = seed / i
    else:
        result = seed

    result = result.quantize(Decimal(str(exp)), rounding=round)

    if isinstance(val, (int, float)):
        return float(result)
    else:
        return result


@mark
def reduce(
    self,
    seed: T,
    op: Union[TH_ASSIGN_OP, Callable[[Any, Any], Any]] = "+=",
    selector=lambda_empty,
) -> T:
    """シーケンスの要素を指定した代入演算子でシードに合成し、合成結果を返す。

    Args:

    * seed: 合成対象とする初期値(左辺)
    * op: 代入演算子または２項演算関数
    * selector: 要素から結合する値を抽出する関数（右辺）


    Usage:
    ```
    >>> pnq.query([1]).reduce(10, "+=")
    11
    >>> pnq.query([[1, 2, 3], [4, 5, 6]]).reduce([], "+=")
    [1, 2, 3, 4, 5, 6]
    >>> pnq.query([{"a": 1}, {"b": 2}]).reduce({}, "|=") # python3.9~
    {"a": 1, "b": 2}
    >>> pnq.query([1, 2, 3, 4, 5]).reduce(0, "+=", lambda x: x * 10)
    150
    >>> pnq.query([1, 2, 3, 4, 5]).reduce(0, lambda l, r: l + r, lambda x: x * 10)
    150
    ```
    """
    if callable(op):
        binary_op = op
    else:
        binary_op = MAP_ASSIGN_OP[op]

    for val in map(selector, self):
        seed = binary_op(seed, val)

    return seed


@mark
def concat(self, selector=lambda x: x, delimiter: str = ""):
    """シーケンスの要素を文字列として連結します。
    Noneは空文字として扱われます。

    Args:

    * selector: 要素から結合する値を抽出する関数
    * delimiter: 区切り文字

    Usage:
    ```
    >>> pnq.query([]).concat()
    ""
    >>> pnq.query([1, 2]).concat()
    "12"
    >>> pnq.query(["a", "b"]).concat()
    "ab"
    >>> pnq.query(["a", None]).concat()
    "a"
    >>> pnq.query(["a", "b"]).concat(delimiter=",")
    "a,b"
    ```
    """
    to_str = lambda x: "" if x is None else str(x)
    return delimiter.join(to_str(x) for x in map(selector, self))


###########################################
# finalizer
###########################################


@mark
def to(self, finalizer):
    """ストリームをファイナライザによって処理します。

    Args:

    * self: 評価するシーケンス
    * finalizer: イテレータを受け取るクラス・関数

    Returns: ファイナライザが返す結果

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).to(list)
    [1, 2]
    >>> pnq.query({1: "a", 2: "b"}).to(dict)
    {1: "a", 2: "b"}
    ```
    """
    if False:  # is_async
        raise Exception()
    else:
        return finalizer(self)


async def to_async(self, finalizer):
    if self.is_async:
        return await finalizer(self)
    else:
        return await finalizer(self)


@mark
def lazy(self, finalizer):
    """ファイナライザの実行するレイジーオブジェクトを返します。
    レイジーオブジェクトをコールすると同期実行され、`await`すると非同期実行されます。

    Args:

    * self: バイパス対象のシーケンス
    * finalizer: イテレータを受け取るクラス・関数

    Returns: ファイナライザが返す結果

    Usage:
    ```
    >>> lazy = pnq.query([1, 2, 3]).lazy(list)
    >>> lazy()
    [1, 2, 3]
    >>> await lazy
    [1, 2, 3]
    >>> lazy = pnq.query([1, 2, 3]).lazy(pnq.actions.first)
    >>> await lazy
    1
    ```
    """
    return finalizer(x for x in self)


@mark
async def to_async(self, cls):
    """非同期ストリームをファイナライザによって処理します。
    クエリが非同期処理を要求する場合のみ使用してください。

    Args:

    * self: バイパス対象のシーケンス
    * finalizer: イテレータを受け取るクラス・関数

    Returns: ファイナライザが返す結果

    Usage:
    ```
    >>> await pnq.query([1, 2, 3]).to_async(list)
    [1, 2]
    >>> await pnq.query({1: "a", 2: "b"}).to_async(dict)
    {1: "a", 2: "b"}
    ```
    """
    return cls(x async for x in self)


@mark
def each(self, func=lambda x: x):
    """イテレーションを実行し、流れてくる要素を関数に送出します。
    例外はコントロールされません。
    関数を指定しない場合、単にイテレーションを実行します。

    Args:

    * self: フィルタ対象のシーケンス
    * func: 値の送出先の関数

    Returns: `None`

    Usage:
    ```
    >>> pnq.query([1,2]).each()
    >>> pnq.query([1,2]).each(print)
    1
    2
    ```
    """
    for elm in self:
        func(elm)


@mark
def each_unpack(self, func=lambda x: x):
    """`each`実行時にキーワードアンパックしながら要素を送出します。
    与える要素は辞書である必要があります。
    基本的な動作は`each`を参照ください。

    Usage:
    ```
    >>> @pnq.query([{"arg1": 1, "arg2": 2}]).each_unpack
    >>> def print_values(arg1, arg2):
    >>>   print(arg1, arg2)
    >>> 1, 2
    ```
    """
    for elm in self:
        func(**elm)


async def async_dummy(*args, **kwargs):
    ...


@mark
async def each_async(self, func=async_dummy):
    """シーケンスから流れてくる値を非同期関数に送出します。
    例外はコントロールされません。

    Args:

    * self: フィルタ対象のシーケンス
    * func: 値の送出先の関数

    Returns: `None`

    Usage:
    ```
    >>> results = []
    >>> async def append(x):
    >>>    results.append(x)
    >>> await pnq.query([1,2]).each_async(append)
    >>> print(results)
    [1, 2]
    ```
    """
    for elm in __iter(self):
        await func(elm)


@mark
async def each_async_unpack(self, func=async_dummy):
    """`each_async`実行時にキーワードアンパックしながら要素を送出します。
    与える要素は辞書である必要があります。
    基本的な動作は`each_async`を参照ください。

    Usage:
    ```
    >>> results = []
    >>> async def append(arg1, arg2):
    >>>    results.append((arg1, arg2))
    >>> await pnq.query([{"arg1": 1, "arg2": 2}]).each_async_unpack(append)
    >>> print(results)
    [(1, 2)]
    ```
    """
    for elm in __iter(self):
        await func(**elm)


@mark
def get(self, key, default=NoReturn):
    """リストや辞書などの`__getitem__`を呼び出します。セットでも使用でき、キーが存在する場合そのままキーを返します。
    デフォルトを設定した場合、キーが存在しない場合はデフォルトを返します。

    このメソッドは実体化している状態でのみ使用することができ、クエリ化されている状態では使用できません。

    Args:

    * key: キー
    * default: キーが存在しない場合に返すデフォルト値

    Usage:
    ```
    >>> data = pnq.query({"a", "b", "c"})
    >>> data.get("a")
    "a"
    >>> data.get("d")
    raise KeyError("d")
    >>> data.get("d", 10)
    10
    ```
    """
    try:
        return self[key]
    except (IndexError, KeyError, NotFoundError):
        pass

    except Exception:
        if isinstance(self, set) and key in self:
            return key

    if default is NoReturn:
        raise NotFoundError(key)
    else:
        return default


@mark
def one(self):
    """シーケンス内の要素が１つであることを検証し、その要素を返します。
    検証に失敗した場合は、例外が発生します。
    デフォルト値を設定した場合は、要素が存在しない場合にデフォルト値を返します。

    `one`関数は、１つの要素であるか検証するために２つ目の要素を取り出そうとします。
    ソースとなるイテラブルが値を消費する実装だと、２つの要素が失われる可能性があることに注意してください。

    Args:

    * default: 要素が存在しない場合に返す値

    Usage:
    ```
    >>> pnq.query([]).one()
    raise NoElementError("...")
    >>> pnq.query([1]).one()
    1
    >>> pnq.query([1, 2]).one()
    raise NotOneElementError("...")
    >>> pnq.query([]).one(None)
    None
    >>> pnq.query([1, 2]).one(None)
    raise NotOneElementError("...")
    ```
    """
    it = iter(self)
    try:
        result = next(it)
    except StopIteration:
        raise NoElementError()

    try:
        next(it)
        raise NotOneElementError()
    except StopIteration:
        pass

    return result


@mark
def first(self):
    """シーケンス内の最初の要素を返します。
    要素が存在しない場合は、例外が発生します。

    セットをソースとした場合、セットは順序を保持しないため、順序性は期待できません。

    Args:

    * default: 要素が存在しない場合に返す値

    Usage:
    ```
    >>> pnq.query([]).first()
    raise NoElementError("...")
    >>> pnq.query([1]).first()
    1
    >>> pnq.query([1, 2]).first()
    1
    >>> pnq.query([]).first(None)
    None
    ```
    """
    # it = __iter(self)
    it = iter(self)
    try:
        return next(it)
    except StopIteration:
        raise NoElementError()


@mark
def last(self):
    """シーケンス内の最後の要素を返します。
    要素が存在しない場合は、例外が発生します。

    セットをソースとした場合、セットは順序を保持しないため、順序性は期待できません。

    Args:

    * default: 要素が存在しない場合に返す値

    Usage:
    ```
    >>> pnq.query([]).last()
    raise NoElementError("...")
    >>> pnq.query([1]).last()
    1
    >>> pnq.query([1, 2]).last()
    2
    >>> pnq.query([]).last(None)
    None
    ```
    """
    if isinstance(self, Sequence):
        try:
            return self[-1]
        except IndexError:
            raise NoElementError()

    undefined = object()
    last = undefined
    for elm in iter(self):
        last = elm

    if last is undefined:
        raise NoElementError()
    else:
        return last


@mark
def get_or(self, key, default):
    try:
        return get(self, key)
    except NotFoundError:
        return default


@mark
def one_or(self, default):
    try:
        return one(self)
    except NoElementError:
        return default


@mark
def first_or(self, default):
    try:
        return first(self)
    except NoElementError:
        return default


@mark
def last_or(self, default):
    try:
        return last(self)
    except NoElementError:
        return default


@mark
def get_or_raise(self, key, exc: Union[str, Exception]):
    """基本的な動作は`get`を参照ください。
    KeyErrorが発生した時、任意の例外を発生させます。

    Args:

    * exc: KeyError時に発生させる例外

    Usage:
    ```
    >>> pnq.query([]).get_or_raise(0, Exception(f"Not Exist Key: 0"))
    raise Exception("Not Exist Key: 0")
    ```
    """
    undefined = object()
    result = get_or(self, key, undefined)
    if result is undefined:
        if isinstance(exc, str):
            raise Exception(exc)
        else:
            raise exc
    else:
        return result


@mark
def one_or_raise(self, exc: Union[str, Exception]):
    """基本的な動作は`one`を参照ください。
    NoElementErrorが発生した時、任意の例外を発生させます。
    NotOneElementErrorはキャッチしません。

    Args:

    * exc: NoElementError時に発生させる例外

    Usage:
    ```
    >>> pnq.query([]).one_or_raise(0, Exception("No exists."))
    raise Exception("No exists.")
    ```
    """
    undefined = object()
    result = one_or(self, undefined)
    if result is undefined:
        if isinstance(exc, str):
            raise Exception(exc)
        else:
            raise exc
    else:
        return result


@mark
def first_or_raise(self, exc: Union[str, Exception]):
    """基本的な動作は`first`を参照ください。
    NoElementErrorが発生した時、任意の例外を発生させます。

    Args:

    * exc: NoElementError時に発生させる例外

    Usage:
    ```
    >>> pnq.query([]).first_or_raise(0, Exception("No exists."))
    raise Exception("No exists.")
    ```
    """
    undefined = object()
    result = first_or(self, undefined)
    if result is undefined:
        if isinstance(exc, str):
            raise Exception(exc)
        else:
            raise exc
    else:
        return result


@mark
def last_or_raise(self, exc: Union[str, Exception]):
    """基本的な動作は`last`を参照ください。
    NoElementErrorが発生した時、任意の例外を発生させます。

    Args:

    * exc: NoElementError時に発生させる例外

    Usage:
    ```
    >>> pnq.query([]).last_or_raise(0, Exception("No exists."))
    raise Exception("No exists.")
    ```
    """
    undefined = object()
    result = last_or(self, undefined)
    if result is undefined:
        if isinstance(exc, str):
            raise Exception(exc)
        else:
            raise exc
    else:
        return result


###########################################
# sleeping
###########################################
def sleep(self, seconds: float):
    from time import sleep as sleep_

    for elm in self:
        yield elm
        sleep_(seconds)


async def sleep_async(self, seconds: float):
    from asyncio import sleep as sleep_

    for elm in self:
        yield elm
        await sleep_(seconds)



import json
from typing import TypeVar

from .base.queries import Query as QueryBase
from .queries import Query

T = TypeVar("T")


class File(Query[T], QueryBase):  # type: ignore
    def __init__(self, *args, **kwargs):
        self.run_iter_type = self.iter_type


class Jsonl(File[dict]):
    def __init__(self, path, deserializer=json.loads):
        super().__init__()
        self.path = path
        self.deserializer = deserializer

    def _impl_iter(self):
        deserializer = self.deserializer
        with open(self.path) as f:
            for line in f.readlines():
                yield deserializer(line)



from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Dict,
    FrozenSet,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    NoReturn,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    no_type_check,
    overload,
)

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from . import actions
from .base import builder, core, finalizers, queries
from .base.exceptions import NoElementError, NotFoundError, NotOneElementError
from .base.op import TH_ASSIGN_OP
from .base.requests import Response

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")
K2 = TypeVar("K2")
V2 = TypeVar("V2")
R = TypeVar("R")


__all__ = ["Query", "PairQuery", "query"]




{% for query in queries %}

class {{query.CLS}}:
    if TYPE_CHECKING:
        def __iter__(self) -> Iterator[{{query.T}}]:
            ...

        def __aiter__(self) -> AsyncIterator[{{query.T}}]:
            ...

    def as_aiter(self) -> "finalizers.AsyncFinalizer[{{query.T}}]":
        return finalizers.AsyncFinalizer(self)

    @property
    def _(self) -> "finalizers.AsyncFinalizer[{{query.T}}]":
        return finalizers.AsyncFinalizer(self)

    async def _call(self):
        return await PnqList.from_aiter(self)
        # return QuerySeq([x async for x in self])

    {% if query.is_pair %}

    def save(self) -> "PnqListPair[{{query.K}}, {{query.V}}]":
        return PnqList(self)

    def __await__(self) -> Generator[Any, Any, "PnqListPair[{{query.K}}, {{query.V}}]"]:
        return self._call().__await__()

    {% else %}

    def save(self) -> "PnqList[{{query.T}}]":
        return PnqList(self)

    def __await__(self) -> Generator[Any, Any, "PnqList[{{query.T}}]"]:
        return self._call().__await__()

    {% endif %}

    def len(self) -> int:
        return actions.len(self)

    def exists(self) -> bool:
        return actions.exists(self)

    def all(self, selector: Callable[[{{query.T}}], Any]=lambda x: x) -> bool:
        return actions.all(self, selector)

    def any(self, selector: Callable[[{{query.T}}], Any]=lambda x: x) -> bool:
        return actions.any(self, selector)

    def contains(self, value, selector: Callable[[{{query.T}}], Any]=lambda x: x) -> bool:
        return actions.contains(self, value, selector)

    @overload
    def min(self, *, default=NoReturn) -> Union[{{query.T}}, NoReturn]: ...

    @overload
    def min(self, selector: Callable[[{{query.T}}], R]=lambda x: x, default=NoReturn) -> R: ...
    def min(self, selector: Callable[[{{query.T}}], R]=lambda x: x, default=NoReturn) -> R:
        return actions.min(self, selector, default)

    @overload
    def max(self, *, default=NoReturn) -> Union[{{query.T}}, NoReturn]: ...
    @overload
    def max(self, selector: Callable[[{{query.T}}], R]=lambda x: x, default=NoReturn) -> R: ...
    def max(self, selector: Callable[[{{query.T}}], R]=lambda x: x, default=NoReturn) -> R:
        return actions.max(self, selector, default)

    @overload
    def sum(self) -> {{query.T}}: ...
    @overload
    def sum(self, selector: Callable[[{{query.T}}], R]=lambda x: x) -> R: ...
    def sum(self, selector: Callable[[{{query.T}}], R]=lambda x: x) -> R:
        return actions.sum(self, selector)

    @overload
    def average(self) -> {{query.T}}: ...
    @overload
    def average(self, selector: Callable[[{{query.T}}], R]=lambda x: x) -> R: ...
    def average(self, selector: Callable[[{{query.T}}], R]=lambda x: x) -> R:
        return actions.average(self, selector)

    def reduce(self, seed: T, op: Union[TH_ASSIGN_OP, Callable[[Any, Any], Any]] = "+=", selector=lambda x: x) -> T:
        return actions.reduce(self, seed, op, selector)

    def concat(self, selector=lambda x: x, delimiter: str = "") -> str:
        return actions.concat(self, selector, delimiter)

    def test(self, expect):
        """同期イテレータと非同期イテレータを実行し結果を比較し、
        それぞれの比較結果(bool)をタプルとして返す。
        実行不可な場合はNoneを返す。
        """

        # TODO: asyncを通るようにパスを切り替えるクエリを挟む必要がある

        import asyncio

        async def to_list():
            return [x async for x in self]

        sync_result = None
        async_result = None

        try:
            sync_result = [x for x in self]
        except Exception as e:
            pass

        try:
            async_result = asyncio.run(to_list())
        except Exception as e:
            pass

        return_sync = None
        return_async = None

        if sync_result is not None:
            return_sync = (sync_result == expect)

        if async_result is not None:
            return_async = (async_result == expect)

        return return_sync, return_async

    {% if query.is_pair %}

    @overload
    def to(self: Iterable[Tuple[K, V]], func: Type[Mapping[K, V]]) -> Mapping[K, V]:
        ...

    @overload
    def to(self: Iterable[Tuple[K, V]], func: Callable[[Iterable[Tuple[K, V]]], R]) -> R:
        ...

    {% endif %}

    @overload
    def to(self: Iterable[T], func: Type[Iterable[T]]) -> Iterable[T]:
        ...

    @overload
    def to(self: Iterable[T], func: Callable[[Iterable[T]], R]) -> R:
        ...

    def to(self, func: Callable[[Iterable[T]], R]) -> R:
        return actions.to(self, func)

    def lazy(self, func, *args, **kwargs):
        return queries.Lazy(self, func, *args, **kwargs)

    def each(self, func: Callable = lambda x: x):
        return actions.each(self, func)

    def each_unpack(self, func: Callable = lambda x: x):
        return actions.each_unpack(self, func)

    async def each_async(self, func: Callable = lambda x: x):
        return await actions.each_async(self, func)

    async def each_async_unpack(self, func: Callable = lambda x: x):
        return await actions.each_async_unpack(self, func)

    def one(self) -> {{query.T}}:
        return actions.one(self)

    def one_or(self, default: R) -> Union[{{query.T}}, R]:
        return actions.one_or(self, default)

    def one_or_raise(self, exc: Union[str, Exception]) -> {{query.T}}:
        return actions.one_or_raise(self, exc)

    def first(self) -> {{query.T}}:
        return actions.first(self)

    def first_or(self, default: R) -> Union[{{query.T}}, R]:
        return actions.first_or(self, default)

    def first_or_raise(self, exc: Union[str, Exception]) -> {{query.T}}:
        return actions.first_or_raise(self, exc)

    def last(self) -> {{query.T}}:
        return actions.last(self)

    def last_or(self, default: R) -> Union[{{query.T}}, R]:
        return actions.last_or(self, default)

    def last_or_raise(self, exc: Union[str, Exception]) -> {{query.T}}:
        return actions.last_or_raise(self, exc)

    @overload
    def cast(self, type: Type[Tuple[K2, V2]]) -> "{{pair.SELF__}}[K2, V2]":
        pass

    @overload
    def cast(self, type: Type[R]) -> "Query[R]":
        pass

    def cast(self, type: Type[R]) -> "Query[R]":
        return self

    def enumerate(self, start: int = 0, step: int = 1) -> "{{pair.SELF__}}[int, {{query.T}}]":
        return queries.Enumerate(self, start, step)

    @overload
    def map(self, selector: Callable[[{{query.T}}], Tuple[K2, V2]]) -> "{{pair.SELF__}}[K2, V2]":
        pass

    @overload
    def map(self, selector: Callable[[{{query.T}}], R]) -> "{{sequence.SELF__}}[R]":
        pass

    def map(self, selector):
        return queries.Map(self, selector)

    {% if query.is_pair %}

    @overload
    def select(self, item: Literal[0]) -> "Query[{{query.K}}]":
        ...

    @overload
    def select(self, item: Literal[1]) -> "Query[{{query.V}}]":
        ...

    {% endif %}

    @overload
    def select(self, field) -> "Query[Any]":
        ...

    @overload
    def select(self, field1, field2, *fields) -> "PairQuery":
        ...

    @overload
    def select(self, field, *fields) -> "Query[Tuple]":
        ...

    def select(self, *fields, attr: bool = False) -> "Query[Any]":
        return queries.Select(self, *fields, attr=attr)

    def select_as_tuple(self, *fields, attr: bool = False) -> "Query[Tuple]":
        return queries.SelectAsTuple(self, *fields, attr=attr)

    def select_as_dict(self, *fields, attr: bool = False, default=NoReturn) -> "Query[Dict]":
        return queries.SelectAsDict(self, *fields, attr=attr, default=default)

    def reflect(self, mapping, default=NoReturn, attr: bool = False):
        return queries.Reflect(self, mapping, attr=attr)

    def flat(self, selector: Callable[..., Iterable[R]] = None) -> "{{sequence.SELF__}}[R]":
        return queries.Flat(self, selector)

    def flat_recursive(self, selector: Callable[[{{query.T}}], Iterable[{{query.T}}]]) -> "{{sequence.SELF__}}[{{query.T}}]":
        return queries.FlatRecursive(self, selector)

    def unpack_pos(self, selector: Callable[..., R]) -> "{{sequence.SELF__}}[R]":
        return queries.UnpackPos(self, selector=selector)

    def unpack_kw(self, selector: Callable[..., R]) -> "{{sequence.SELF__}}[R]":
        return queries.UnpackKw(self, selector=selector)

    def pivot_unstack(self, default=None) -> "PairQuery[Any, List]":
        return queries.PivotUnstack(self, default=default)

    def pivot_stack(self) -> "Query[Dict]":
        return queries.PivotStack(self)

    def group_by(self, selector: Callable[[{{query.T}}], Tuple[K2, V2]] = lambda x: x) -> "{{pair.SELF__}}[K2, List[V2]]":
        return queries.GroupBy(self, selector=selector)

    def chunked(self, size: int) -> "Query[List[{{query.T}}]]":
        return queries.Chunked(self, size=size)

    def tee(self, size: int):
        return queries.Tee(self, size=size)

    def join(self, right, on: Callable[[Tuple[list, list]], Callable], select):
        [].join(
            [],
            lambda left, right: left.name == right.name,
            lambda left, right: (left.name, right.age)
        )

        table(User).join(
            Item,
            on=User.id == Item.id
        ).select(User.id, Item.id)

        pass

    def request(self, func, retry: int = None) -> "Query[Response]":
        return queries.Request(self, func, retry)

    def request_async(self, func, retry: int = None, timeout=None) -> "Query[Response]":
        return queries.RequestAsync(self, func, retry=retry, timeout=None)

    def debug(self, breakpoint=lambda x: x, printer=print) -> "{{query.SELF_T}}":
        return queries.Debug(self, breakpoint=breakpoint, printer=printer)

    def debug_path(self, selector_sync=lambda x: -10, selector_async=lambda x: 10) -> "{{query.SELF_T}}":
        return queries.DebugPath(self, selector_sync, selector_async)

    def filter(self, predicate: Callable[[{{query.T}}], bool]) -> "{{query.SELF_T}}":
        return queries.Filter(self, predicate)

    def filter_type(self, *types: Type[R]) -> "{{query.SELF_T}}":
        return queries.FilterType(self, *types)

    @overload
    def filter_unique(self) -> "{{query.SELF_T}}":
        ...

    @overload
    def filter_unique(self, selector: Callable[[{{query.T}}], Tuple[K2, V2]]) -> "{{pair.SELF__}}[K2, V2]":
        ...

    @overload
    def filter_unique(self, selector: Callable[[{{query.T}}], R]) -> "{{sequence.SELF__}}[R]":
        ...

    def filter_unique(self, selector=None):
        return queries.FilterUnique(self, selector=selector)

    def distinct(self, selector: Callable[[{{query.T}}], Any]) -> "{{query.SELF_T}}":
        return queries.FilterUnique(self, selector=selector)

    def must(self, predicate: Callable[[{{query.T}}], bool], msg: str="") -> "{{query.SELF_T}}":
        return queries.Must(self, predicate, msg)

    def must_type(self, type, *types: Type) -> "{{query.SELF_T}}":
        return queries.MustType(self, type, *types)

    def must_unique(self, selector: Callable[[T], R] = None):
        return queries.MustUnique(self, selector=selector)

    def take(self, count_or_range: Union[int, range]) -> "{{query.SELF_T}}":
        return queries.Take(self, count_or_range)

    def take_while(self, predicate) -> "{{query.SELF_T}}":
        return queries.TakeWhile(self, predicate)

    def skip(self, count_or_range: Union[int, range]) -> "{{query.SELF_T}}":
        return queries.Skip(self, count_or_range)

    def take_page(self, page: int, size: int) -> "{{query.SELF_T}}":
        return queries.TakePage(self, page=page, size=size)

    def order_by(self, *fields, desc: bool = False, attr: bool = False) -> "{{query.SELF_T}}":
        return queries.OrderBy(self, *fields, desc=desc, attr=attr)

    def order_by_map(self, selector=None, *, desc: bool = False) -> "{{query.SELF_T}}":
        return queries.OrderByMap(self, selector=selector, desc=desc)

    def order_by_reverse(self) -> "{{query.SELF_T}}":
        return queries.OrderByReverse(self)

    def order_by_shuffle(self) -> "{{query.SELF_T}}":
        return queries.OrderByShuffle(self)

    def sleep(self, seconds: float) -> "{{query.SELF_T}}":
        return queries.Sleep(self, seconds)

    def sleep_async(self, seconds: float) -> "{{query.SELF_T}}":
        return queries.Sleep(self, seconds)

    def zip(self):
        raise NotImplementedError()

    def cartesian(self, *iterables) -> "Query[Tuple]":
        return queries.Cartesian(self, *iterables)

{% endfor %}


if not TYPE_CHECKING:
    import types

    class Queries:
        pass

    from .base import queries

    classess = Queries()

    for cls in queries.exports:
        baseclasses = (cls, Query[T])
        created = types.new_class(cls.__name__, baseclasses)

        setattr(classess, cls.__name__, created)

    queries = classess


class QueryBase(Query[T], core.Query[T]):
    pass


class QueryAsync(Query[T], core.QueryAsync[T]):
    pass


class QueryNormal(Query[T], core.QueryNormal[T]):
    pass


class QueryDict(PairQuery[K, V], core.QueryDict[K, V]):
    def filter_keys(self, *keys) -> "PairQuery[K, V]":
        return queries.FilterKeys(self, *keys)

    # @no_type_check
    def must_keys(self, *keys) -> "PairQuery[K, V]":
        return queries.MustKeys(self, *keys, typ="map")

    def get(self, key) -> V:
        try:
            return self.source[key]
        except KeyError:
            raise NotFoundError(key)

    def get_or(self, key, default: R = None) -> Union[V, R]:
        try:
            return self.get(key)
        except NotFoundError:
            return default

    def get_or_raise(self, key, exc: Union[str, Exception]) -> V:
        undefined = object()
        result = self.get_or(key, undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result


class QuerySeq(Query[T], core.QuerySeq[T]):
    def filter_keys(self, *keys) -> "Query[T]":
        return queries.FilterKeys(self, *keys)

    def must_keys(self, *keys) -> "Query[T]":
        return queries.MustKeys(self, *keys, typ="seq")

    def get(self, key) -> T:
        try:
            return self.source[key]
        except IndexError:
            raise NotFoundError(key)

    def get_or(self, key, default: R = None) -> Union[T, R]:
        try:
            return self.get(key)
        except NotFoundError:
            return default

    def get_or_raise(self, key, exc: Union[str, Exception]) -> T:
        undefined = object()
        result = self.get_or(key, undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result


class PnqList(Query[T], List[T]):
    @property
    def source(self):
        return self

    @classmethod
    async def from_aiter(cls, aiter: AsyncIterable[T]) -> "PnqList[T]":
        result = cls()
        async for x in aiter:
            result.append(x)

        return result

    def filter_keys(self, *keys) -> "Query[T]":
        return queries.FilterKeys(self, *keys)

    def must_keys(self, *keys) -> "Query[T]":
        return queries.MustKeys(self, *keys, typ="seq")

    def get(self, key) -> T:
        try:
            return self.source[key]
        except IndexError:
            raise NotFoundError(key)

    def get_or(self, key, default: R = None) -> Union[T, R]:
        try:
            return self.get(key)
        except NotFoundError:
            return default

    def get_or_raise(self, key, exc: Union[str, Exception]) -> T:
        undefined = object()
        result = self.get_or(key, undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result


class QuerySet(Query[T], core.QuerySet[T]):
    def filter_keys(self, *keys) -> "Query[T]":
        return queries.FilterKeys(self, *keys)

    def must_keys(self, *keys) -> "Query[T]":
        return queries.MustKeys(self, *keys, typ="set")

    def get(self, key) -> T:
        if key in self.source:
            return key
        else:
            raise NotFoundError(key)

    def get_or(self, key, default: R = None) -> Union[T, R]:
        try:
            return self.get(key)
        except NotFoundError:
            return default

    def get_or_raise(self, key, exc: Union[str, Exception]) -> T:
        undefined = object()
        result = self.get_or(key, undefined)
        if result is undefined:
            if isinstance(exc, str):
                raise Exception(exc)
            else:
                raise exc
        else:
            return result

if TYPE_CHECKING:
    class QuerySeqPair(PairQuery[K, V], core.QuerySeq[Tuple[K, V]]):
        ...

    class QuerySetPair(PairQuery[K, V], core.QuerySeq[Tuple[K, V]]):
        ...

    class PnqListPair(PairQuery[K, V], PnqList[Tuple[K, V]]):
        def __init__(self, source: Iterable[Tuple[K, V]]):
            ...


class QueryBuilder(builder.Builder):
    QUERY_BOTH = QueryBase
    QUERY_ASYNC = QueryAsync
    QUERY_NORMAL = QueryNormal
    QUERY_SEQ = QuerySeq
    QUERY_DICT = QueryDict
    QUERY_SET = QuerySet


@overload
def query(source: Mapping[K, V]) -> QueryDict[K, V]:
    ...


@overload
def query(source: Set[Tuple[K, V]]) -> "QuerySetPair[K, V]":
    ...

@overload
def query(source: Set[T]) -> QuerySet[T]:
    ...

@overload
def query(source: Iterable[Tuple[K, V]]) -> "QuerySeqPair[K, V]":
    ...

@overload
def query(source: Iterable[T]) -> QuerySeq[T]:
    ...

@overload
def query(source: AsyncIterable[Tuple[K, V]]) -> "QuerySeqPair[K, V]":
    ...

@overload
def query(source: AsyncIterable[T]) -> QuerySeq[T]:
    ...


@overload
def query(source) -> "Query[Any]":
    ...


def query(source):
    return QueryBuilder.query(source)


run = QueryBuilder.run



# isort: off

# generating
# transforming
from .base.actions import debug as debug
from .base.actions import unpack_kw as unpack_kw
from .base.actions import unpack_pos as unpack_pos
from .base.actions import flat as flat
from .base.actions import flat_recursive as flat_recursive
from .base.actions import pivot_unstack as pivot_unstack
from .base.actions import pivot_stack as pivot_stack
from .base.actions import reflect as reflect

from .base.actions import chunked as chunked
from .base.actions import tee as tee


# filtering
from .base.actions import filter_unique as filter_unique
from .base.actions import filter_keys as filter_keys

# validating
# sorting
# partitioning
from .base.actions import skip as skip
from .base.actions import skip_while as skip_while
from .base.actions import take as take
from .base.actions import take_page as take_page
from .base.actions import take_while as take_while

# expanding
# executing
from .base.actions import each as each
from .base.actions import each_unpack as each_unpack
from .base.actions import each_async as each_async
from .base.actions import each_async_unpack as each_async_unpack

# aggregating
# getting


from .base.actions import __all as all
from .base.actions import __any as any
from .base.actions import __enumerate as enumerate
from .base.actions import __filter as filter
from .base.actions import __iter as iter
from .base.actions import __len as len
from .base.actions import __map as map
from .base.actions import __max as max
from .base.actions import __min as min
from .base.actions import __range as range
from .base.actions import __sum as sum
from .base.actions import __zip as zip
from .base.actions import compress as compress
from .base.actions import average as average
from .base.actions import cast as cast
from .base.actions import concat as concat
from .base.actions import contains as contains
from .base.actions import count as count
from .base.actions import cycle as cycle

from .base.actions import __list as list
from .base.actions import __dict as dict

from .base.actions import distinct as distinct
from .base.actions import exists as exists
from .base.actions import filter_type as filter_type
from .base.actions import first as first
from .base.actions import first_or as first_or
from .base.actions import first_or_raise as first_or_raise
from .base.actions import get as get
from .base.actions import get_or as get_or
from .base.actions import get_or_raise as get_or_raise
from .base.actions import group_by as group_by
from .base.actions import group_join as group_join
from .base.actions import infinite as infinite
from .base.actions import join as join
from .base.actions import last as last
from .base.actions import last_or as last_or
from .base.actions import last_or_raise as last_or_raise
from .base.actions import lazy as lazy

from .base.actions import must as must
from .base.actions import must_keys as must_keys

from .base.actions import must_type as must_type
from .base.actions import must_unique as must_unique
from .base.actions import one as one
from .base.actions import one_or as one_or
from .base.actions import one_or_raise as one_or_raise
from .base.actions import order_by as order_by
from .base.actions import order_by_fields as order_by_fields
from .base.actions import order_by_reverse as order_by_reverse
from .base.actions import order_by_shuffle as order_by_shuffle
from .base.actions import reduce as reduce
from .base.actions import repeat as repeat
from .base.actions import request as request
from .base.actions import request_async as request_async
from .base.actions import select as select
from .base.actions import select_as_dict as select_as_dict
from .base.actions import select_as_tuple as select_as_tuple
from .base.actions import select_single_node as select_single_nod
from .base.actions import to as to
from .base.actions import union as union
from .base.actions import union_all as union_all
from .base.actions import union_intersect as union_intersect
from .base.actions import union_minus as union_minus

from .base.actions import value as value

# sleep
from .base.actions import sleep as sleep
from .base.actions import sleep_async as sleep_async

# other
from .base.actions import raise_if_not_unique_keys as raise_if_not_unique_keys
from .base.actions import take_page_calc as take_page_calc

# isort: on



import asyncio
import itertools
from collections import defaultdict
from operator import attrgetter, itemgetter
from typing import Awaitable, Callable, Mapping, NoReturn, TypeVar, Union

from .core import IterType, Query, QueryDict, QuerySeq, QuerySet
from .exceptions import (
    DuplicateElementError,
    MustError,
    MustTypeError,
    NoElementError,
    NotFoundError,
    NotOneElementError,
)

R = TypeVar("R")

exports = []


def mark(cls):
    exports.append(cls)
    return cls


@mark
class Lazy(Query, Awaitable[R]):
    def __init__(self, source, finalizer: Callable[..., R], *args, **kwargs):
        super().__init__(source)
        self.finalizer = finalizer
        self.args = args
        self.kwargs = kwargs

    def __call__(self) -> R:
        return self.finalizer(self.__iter__(), *self.args, **self.kwargs)

    def __await__(self):
        coro = self.finalizer(self.__aiter__(), *self.args, **self.kwargs)
        return coro.__await__()


@mark
class AsyncMap(Query):
    iter_type = IterType.ASYNC

    def __init__(self, source, selector):
        super().__init__(source)
        self.selector = selector

    async def _impl_aiter(self):
        selector = self.selector
        async for x in self.source:
            yield await selector(x)


@mark
class Sleep(Query):
    iter_type = IterType.ASYNC

    def __init__(self, source, seconds):
        super().__init__(source)
        self.seconds = seconds

    async def _impl_aiter(self):
        seconds = self.seconds
        sleep = asyncio.sleep

        async for v in self.source:
            yield v
            await sleep(seconds)


@mark
class Map(Query):
    def __init__(self, source, selector):
        super().__init__(source)
        if selector is str:
            selector = lambda x: "" if x is None else str(x)
        self.selector = selector

    def _impl_iter(self):
        selector = self.selector
        if selector is None:
            return self.source.__iter__()
        else:
            return map(selector, self.source)

    def _impl_aiter(self):
        selector = self.selector
        if selector is None:
            return self.source.__aiter__()
        else:
            return (selector(x) async for x in self.source)


@mark
class UnpackPos(Map):
    def _impl_iter(self):
        selector = self.selector
        return (selector(*v) for v in self.source)

    def _impl_aiter(self):
        selelctor = self.selector
        return (selelctor(*v) async for v in self.source)


@mark
class UnpackKw(Map):
    def _impl_iter(self):
        selector = self.selector
        return (selector(**v) for v in self.source)

    def _impl_aiter(self):
        selector = self.selector
        return (selector(**v) async for v in self.source)


@mark
class Select(Map):
    def __init__(self, source, *fields, attr: bool = False):
        if attr:
            selector = attrgetter(*fields)
        else:
            selector = itemgetter(*fields)
        super().__init__(source, selector)


@mark
class SelectAsTuple(Map):
    def __init__(self, source, *fields, attr: bool = False):
        if len(fields) > 1:
            if attr:
                selector = attrgetter(*fields)
            else:
                selector = itemgetter(*fields)
        elif len(fields) == 1:
            field = fields[0]
            if attr:
                selector = lambda x: (getattr(x, field),)
            else:
                selector = lambda x: (x[field],)
        else:
            selector = lambda x: ()
        super().__init__(source, selector)


@mark
class SelectAsDict(Map):
    def __init__(self, source, *fields, attr: bool = False, default=NoReturn):
        if attr:
            if default is NoReturn:
                selector = lambda x: {k: getattr(x, k) for k in fields}  # noqa
            else:
                selector = lambda x: {k: getattr(x, k, default) for k in fields}  # noqa
        else:
            if default is NoReturn:
                selector = lambda x: {k: x[k] for k in fields}  # noqa
            else:

                def get(obj, k):
                    try:
                        return obj[k]
                    except Exception:
                        return default

                selector = lambda x: {k: get(x, k) for k in fields}  # noqa
        super().__init__(source, selector)


def transpose(mapping):

    tmp = defaultdict(list)

    for left, right in mapping.items():
        if isinstance(right, str):
            tmp[left].append(right)
        elif isinstance(right, list):
            tmp[left] = right
        elif isinstance(right, tuple):
            tmp[left] = right
        elif isinstance(right, set):
            tmp[left] = right
        else:
            raise TypeError(f"{v} is not a valid mapping")

    # output属性 - 元の属性（複数の場合あり）
    target = defaultdict(list)

    for k, outputs in tmp.items():
        for out in outputs:
            target[out].append(k)

    return target


def split_single_multi(dic):
    single = {}
    multi = {}
    for k, v in dic.items():
        if len(v) > 1:
            multi[k] = v
        else:
            single[k] = v[0]

    return single, multi


def build_selector(single, multi, attr: bool = False):
    template = {}
    for k in multi.keys():
        template[k] = []

    if attr:

        def reflector(x):
            result = {}
            for k, v in single.items():
                result[k] = x[v]

            for k, fields in multi.items():
                result[k] = []
                for f in fields:
                    result[k].append(x[f])

            return result

    else:

        def reflector(x):
            result = {}
            for k, v in single.items():
                result[k] = x[v]

            for k, fields in multi.items():
                result[k] = []
                for f in fields:
                    result[k].append(x[f])

            return result

    return reflector


@mark
class Reflect(Map):
    def __init__(self, source, mapping, attr: bool = False):
        transposed = transpose(mapping)
        single, multi = split_single_multi(transposed)
        selector = build_selector(single, multi, attr)
        super().__init__(source, selector)


@mark
class Flat(Query):
    def __init__(self, source, selector=None):
        super().__init__(source)
        self.selector = selector

    def _impl_iter(self):
        selector = self.selector
        if selector is None:
            return (_ for inner in self.source for _ in inner)
        else:
            return (_ for elm in self.source for _ in selector(elm))

    def _impl_aiter(self):
        selector = self.selector
        if selector is None:
            return (_ async for inner in self.source async for _ in inner)
        else:
            return (_ async for elm in self.source async for _ in selector(elm))


@mark
class FlatRecursive(Query):
    def __init__(self, source, selector):
        super().__init__(source)

        def scan(parent, selector):
            nodes = selector(parent)
            if nodes is None:
                return
            for node in nodes:
                yield node
                yield from scan(node, selector)

        self.scan = scan
        self.selector = selector

    def _impl_iter(self):
        scan = self.scan
        selector = self.selector
        for node in self.source:
            yield node
            yield from scan(node, selector)

    async def _impl_aiter(self):
        scan = self.scan
        selector = self.selector
        async for node in self.source:
            yield node
            for c in scan(node, selector):
                yield c


@mark
class Enumerate(Query):
    def __init__(self, source, start: int = 0, step: int = 1):
        super().__init__(source)
        self.start = start
        self.step = step

    def _impl_iter(self):
        step = self.step
        return (x * step for x in enumerate(self.source, self.start))

    async def _impl_aiter(self):
        step = self.step
        i = self.start - step
        async for x in self.source:
            i += step
            yield i, x


@mark
class PivotUnstack(Query):
    def __init__(self, source, default=None) -> None:
        super().__init__(source)
        self.default = default

    def _impl_iter(self):
        default = self.default
        dataframe = {}  # type: ignore
        data = []

        # 全てのカラムを取得
        for i, dic in enumerate(self.source):
            data.append(dic)
            for k in dic.keys():
                dataframe[k] = None

        # カラム分の領域を初期化
        for k in dataframe:
            dataframe[k] = []

        # データをデータフレームに収める
        for dic in data:
            for k in dataframe.keys():
                v = dic.get(k, default)
                dataframe[k].append(v)

        yield from dataframe.items()

    async def _impl_aiter(self):
        raise NotImplementedError()


@mark
class PivotStack(Query):
    def _impl_iter(self):
        data = dict(self.source.__iter__())
        columns = list(data.keys())

        for i in range(len(columns)):
            row = {}
            for c in columns:
                row[c] = data[c][i]

            yield row

    async def _impl_aiter(self):
        raise NotImplementedError()


@mark
class Request(Query):
    iter_type = IterType.BOTH

    def __init__(self, source, func, retry: int = 0) -> None:
        super().__init__(source)
        self.func = func
        self.retry = retry

        if retry:
            raise NotImplementedError("retry not implemented")

    def _impl_iter(self):
        func = self.func

        from .requests import Response, StopWatch

        for v in self.source:

            with StopWatch() as sw:
                err = None
                result = None
                try:
                    result = func(**v)
                except Exception as e:
                    err = e

            res = Response(
                func, kwargs=v, err=err, result=result, start=sw.start, end=sw.end
            )

            yield res


@mark
class RequestAsync(Query):
    iter_type = IterType.ASYNC

    def __init__(self, source, func, timeout: float = None, retry: int = 0) -> None:
        super().__init__(source)
        self.func = func

        if retry:
            raise NotImplementedError("retry not implemented")

        if timeout:
            raise NotImplementedError("timeout not implemented")

    def _impl_iter(self):
        raise NotImplementedError()

    async def _impl_aiter(self):
        func = self.func

        from .requests import Response, StopWatch

        async for v in self.source:

            with StopWatch() as sw:
                err = None
                result = None
                try:
                    result = await func(**v)
                except Exception as e:
                    err = e

            res = Response(
                func, kwargs=v, err=err, result=result, start=sw.start, end=sw.end
            )

            yield res


@mark
class Debug(Query):
    def __init__(self, source, breakpoint=lambda x: x, printer=print) -> None:
        super().__init__(source)
        self.breakpoint = breakpoint
        self.printer = printer

    def _impl_iter(self):
        breakpoint = self.breakpoint
        printer = self.printer

        for v in self.source:
            printer(v)
            breakpoint(v)
            yield v

    async def _impl_aiter(self):
        breakpoint = self.breakpoint
        printer = self.printer

        async for v in self.source:
            printer(v)
            breakpoint(v)
            yield v


@mark
class DebugPath(Query):
    """同期イテレータと非同期イテレータのどちらが実行されているか確認するデバッグ用のクエリです。"""

    def __init__(
        self, source, selector_sync=lambda x: -10, selector_async=lambda x: 10
    ) -> None:
        super().__init__(source)
        self.selector_sync = selector_sync
        self.selector_async = selector_async

    def _impl_iter(self):
        selector_sync = self.selector_sync
        for v in self.source:
            yield selector_sync(v)

    async def _impl_aiter(self):
        selector_async = self.selector_async
        async for v in self.source:
            yield selector_async(v)


@mark
class GroupBy(Query):
    def __init__(self, source, selector):
        super().__init__(source)
        self.selector = selector

    def _impl_iter(self):
        selector = self.selector
        results = defaultdict(list)
        for elm in self.source:
            k, v = selector(elm)
            results[k].append(v)

        for k, v in results.items():
            yield k, v

    async def _impl_aiter(self):
        selector = self.selector
        results = defaultdict(list)
        async for elm in self.source:
            k, v = selector(elm)
            results[k].append(v)

        for k, v in results.items():
            yield k, v


@mark
class Chunked(Query):
    def __init__(self, source, size: int) -> None:
        super().__init__(source)
        self.size = size

        if size <= 0:
            raise ValueError("count must be greater than 0")

    def _impl_iter(self):
        size = self.size
        current = 0
        running = True

        it = iter(self.source)

        while running:
            current = 0
            queue = []

            while current < size:
                current += 1
                try:
                    val = next(it)
                    queue.append(val)
                except StopIteration:
                    running = False
                    break

            if queue:
                yield queue

    async def _impl_aiter(self):
        size = self.size
        current = 0
        running = True

        it = self.source.__aiter__()

        while running:
            current = 0
            queue = []

            while current < size:
                current += 1
                try:
                    val = await it.__anext__()
                    queue.append(val)
                except StopIteration:
                    running = False
                    break

            if queue:
                yield queue


@mark
class Tee(Query):
    def __init__(self, source, size: int = 2) -> None:
        super().__init__(source)
        self.size = size

    def _impl_iter(self):
        return iter(itertools.tee(self.source, self.size))


@mark
class Filter(Query):
    def __init__(self, source, predicate):
        super().__init__(source)
        self.predicate = predicate

    def _impl_iter(self):
        return filter(self.predicate, self.source)

    async def _impl_aiter(self):
        predicate = self.predicate
        async for x in self.source:
            if predicate(x):
                yield x


@mark
class Must(Query):
    def __init__(self, source, predicate, msg: str = ""):
        super().__init__(source)
        self.predicate = predicate
        self.msg = msg

    def _impl_iter(self):
        predicate = self.predicate
        for elm in self.source:
            if not predicate(elm):
                raise MustError(f"{self.msg} {elm}")
            yield elm

    async def _impl_aiter(self):
        predicate = self.predicate
        async for elm in self.source:
            if not predicate(elm):
                raise MustError(f"{self.msg} {elm}")
            yield elm


@mark
class FilterType(Query):
    def __init__(self, source, *types):
        super().__init__(source)
        self.types = tuple(None.__class__ if x is None else x for x in types)

    def _impl_iter(self):
        types = self.types
        for elm in self.source:
            if isinstance(elm, types):
                yield elm

    async def _impl_aiter(self):
        types = self.types
        async for elm in self.source:
            if isinstance(elm, types):
                yield elm


@mark
class MustType(Query):
    def __init__(self, source, type, *types):
        super().__init__(source)
        self.types = (type, *types) if types else type

    def _impl_iter(self):
        types = self.types
        for elm in self.source:
            if not isinstance(elm, types):
                raise MustTypeError(f"{elm} is not {types}")
            yield elm

    async def _impl_aiter(self):
        types = self.types
        async for elm in self.source:
            if not isinstance(elm, types):
                raise MustTypeError(f"{elm} is not {types}")
            yield elm


@mark
class FilterUnique(Query):
    def __init__(self, source, selector=None):
        if selector is not None:
            source = Map(source, selector)
        super().__init__(source)

    def _impl_iter(self):
        duplidate = set()
        for value in self.source:
            if value in duplidate:
                pass
            else:
                duplidate.add(value)
                yield value

    async def _impl_aiter(self):
        duplidate = set()
        async for value in self.source:
            if value in duplidate:
                pass
            else:
                duplidate.add(value)
                yield value


@mark
class MustUnique(Query):
    def __init__(self, source, selector=None):
        super().__init__(source)

        if selector is not None:
            self.selector = selector
        else:
            self.selector = lambda x: x

    def _impl_iter(self):
        selector = self.selector
        duplidate = set()
        for value in self.source:
            target = selector(value)
            if target in duplidate:
                raise DuplicateElementError(value)
            else:
                duplidate.add(target)
                yield value

    async def _impl_aiter(self):
        selector = self.selector
        duplidate = set()
        async for value in self.source:
            target = selector(value)
            if target in duplidate:
                raise DuplicateElementError(value)
            else:
                duplidate.add(target)
                yield value


def get_many_for_mapping(query_dict, keys):
    """"""
    undefined = object()
    for key in keys:
        obj = query_dict.get_or(key, undefined)
        if obj is not undefined:
            yield key, obj


def get_many_for_sequence(query_seq, keys):
    """"""
    undefined = object()
    for key in keys:
        obj = query_seq.get_or(key, undefined)
        if obj is not undefined:
            yield obj


def get_many_for_set(query_set, keys):
    """"""
    undefined = object()
    for key in keys:
        obj = query_set.get_or(key, undefined)
        if obj is not undefined:
            yield key


def get_duplicate_keys(keys):
    if len(keys) == len(set(keys)):
        return None

    from collections import Counter

    dup = []
    for k, count in Counter(keys).items():  # type: ignore
        if count > 1:
            dup.append(k)

    return dup


@mark
class FilterKeys(Query):
    def __init__(self, source, *keys):
        super().__init__(source)
        self.keys = keys

        dup = get_duplicate_keys(keys)
        if dup:
            raise ValueError("duplicate keys: {dup}")

        if isinstance(source, QueryDict):
            self.filter = get_many_for_mapping
        elif isinstance(source, QuerySeq):
            self.filter = get_many_for_sequence
        elif isinstance(source, QuerySet):
            self.filter = get_many_for_set
        else:
            raise TypeError(f"{source} is not QuerySeq, QueryDict or QuerySet")

    def _impl_iter(self):
        return self.filter(self.source, self.keys)

    async def _impl_aiter(self):
        for x in self._impl_iter():
            yield x


@mark
class MustKeys(Query):
    # Literal["set", "seq", "map"]
    def __init__(self, source, *keys, typ: str):
        super().__init__(source)
        self.keys = keys
        self.type = typ

        dup = get_duplicate_keys(keys)
        if dup:
            raise ValueError("duplicate keys: {dup}")

    def _impl_iter(self):
        source = self.source
        not_exists = set()
        key_values = []
        undefined = object()

        for k in self.keys:
            val = source.get_or(k, undefined)
            if val is undefined:
                not_exists.add(k)
            else:
                key_values.append((k, val))

        if not_exists:
            raise NotFoundError(str(not_exists))

        if self.type == "map":
            for k, v in key_values:
                yield k, v
        elif self.type == "seq":
            for k, v in key_values:
                yield v
        elif self.type == "set":
            for k, v in key_values:
                yield k
        else:
            raise TypeError(f"unknown type: {self.type}")


@mark
class Take(Query):
    def __init__(self, source, count_or_range: Union[int, range]):
        super().__init__(source)
        if isinstance(count_or_range, range):
            r = count_or_range
        else:
            r = range(count_or_range)
        self.start = r.start
        self.stop = r.stop

        if self.start < 0:
            self.start = 0

        if self.stop < 0:
            self.stop = 0

    def _impl_iter(self):
        return itertools.islice(self.source, self.start, self.stop)

    async def _impl_aiter(self):
        start = self.start
        stop = self.stop

        current = 0

        it = self.source.__aiter__()

        try:
            while current < start:
                await it.__anext__()
                current += 1
        except StopAsyncIteration:
            return

        try:
            while current < stop:
                yield await it.__anext__()
                current += 1
        except StopAsyncIteration:
            return


@mark
class TakeWhile(Query):
    def __init__(self, source, predicate):
        super().__init__(source)
        self.predicate = predicate

    def _impl_iter(self):
        predicate = self.predicate
        for v in self.source:
            if predicate(v):
                yield v
            else:
                break

    async def _impl_aiter(self):
        predicate = self.predicate
        async for v in self.source:
            if predicate(v):
                yield v
            else:
                break


@mark
class Skip(Query):
    def __init__(self, source, count_or_range: Union[int, range]):
        super().__init__(source)

        if isinstance(count_or_range, range):
            r = count_or_range
        else:
            r = range(count_or_range)
        self.start = r.start
        self.stop = r.stop

    def _impl_iter(self):
        start = self.start
        stop = self.stop

        current = 0

        it = iter(self.source)

        try:
            while current < start:
                yield next(it)
                current += 1
        except StopIteration:
            return

        try:
            while current < stop:
                next(it)
                current += 1
        except StopIteration:
            return

        for x in it:
            yield x

    async def _impl_aiter(self):
        start = self.start
        stop = self.stop

        current = 0

        it = self.source.__aiter__()

        try:
            while current < start:
                yield await it.__anext__()
                current += 1
        except StopAsyncIteration:
            return

        try:
            while current < stop:
                await it.__anext__()
                current += 1
        except StopAsyncIteration:
            return

        async for x in it:
            yield x


def take_page_calc(page: int, size: int):
    if size < 0:
        raise ValueError("size must be >= 0")
    start = (page - 1) * size
    stop = start + size
    return start, stop


@mark
class TakePage(Take):
    def __init__(self, source, page: int, size: int):
        start, stop = take_page_calc(page, size)
        super().__init__(source, range(start, stop))


@mark
class OrderByMap(Query):
    def __init__(self, source, selector=None, desc: bool = False) -> None:
        super().__init__(source)
        self.selector = selector
        self.desc = desc

    def _impl_iter(self):
        yield from sorted(self.source, key=self.selector, reverse=self.desc)

    async def _impl_aiter(self):
        raise NotImplementedError()


@mark
class OrderBy(OrderByMap):
    def __init__(self, source, *fields, desc: bool = False, attr: bool = False) -> None:
        if not len(fields):
            selector = None
        else:
            if attr:
                selector = attrgetter(*fields)
            else:
                selector = itemgetter(*fields)

        super().__init__(source, selector=selector, desc=desc)


@mark
class OrderByReverse(Query):
    def _impl_iter(self):
        source = self.source
        if hasattr(source, "__reversed__"):
            if isinstance(source, Mapping):
                return ((k, source[k]) for k in reversed(source))  # type: ignore
            else:
                return reversed(source)
        else:
            return reversed(list(source))

    def _impl_aiter(self):
        raise NotImplementedError()


@mark
class OrderByShuffle(OrderByMap):
    def _impl_iter(self):
        import random

        yield from sorted(self.source, key=lambda k: random.random())

    def _impl_aiter(self):
        raise NotImplementedError()


@mark
class Cartesian(Query):
    def __init__(self, source, *iterables):
        super().__init__(source)
        self.iterables = iterables

    def _impl_iter(self):
        import itertools

        return itertools.product(self.source, *self.iterables)

    def _impl_aiter(self):
        raise NotImplementedError()
